#include <xs1.h>
#include <platform.h>
#include <xscope.h>
#include <stdlib.h>
#include <math.h>

#include "mpu6050.h"
#include "motor_control.h"
#include "pid.h"
#include "conio.h"
#include "leds.h"
#include "uart.h"

struct IMU imu = {{on tile[0]:XS1_PORT_1P, on tile[0]:XS1_PORT_1O}};

#define IMU_INTERVAL     800000
#define LOGSIZE          300
#define TICKS_PER_SECOND 100000000
#define BUFFER_LENGTH    3
#define OFFSET           0.425
#define GET_DUMP         0
#define MAX_ANGLE        10.0

enum Request{UPDATE, DUMP};

typedef struct {
    float elapsed;
    float estimate;
    int   response;
} robot_state;

void angle_sensor(chanend to_controller) {
    float gx, gy, gz;
    float ax, ay;
    float estimated ;
    timer tmr;
    int t1;

    imu_setup(imu);

    if (mpu_read_byte(imu.i2c, MPU6050_RA_WHO_AM_I) != 0x68) {
        while(1);
    }
    calibrate_gyros(imu);

    tmr :> t1;
    while(1) {
        {gx, gy, gz} = get_gyro(imu);
        {ax, ay}     = get_accel_angles(imu);

        estimated = estimate_angle(gz, ax, IMU_INTERVAL);
        to_controller <: IMU_INTERVAL;
        to_controller <: estimated;
        t1 += IMU_INTERVAL;
        tmr when timerafter(t1) :> void;
    }
}

void controller(chanend from_imu,
                chanend from_motor, chanend to_motor,
                chanend from_io, chanend to_io, chanend to_leds)
{
    robot_state state = {0, 0.0, 0};
    robot_state log[LOGSIZE];
    pid_state pstate = {228.0, 4145.0, 1.25, 150, -150};

    int   delta_ticks;
    float delta_t;
    int   io_request;
    int   log_index = 0;
    int   dont_stop = 1;
    int   error_flag  = 0;
    float estimate;
    float accumulator ;

    float angle_buffer[BUFFER_LENGTH];
    int   buffer_index = 0;

    while(dont_stop) {
        select {
            case from_io :> io_request:
                if(io_request == UPDATE) {
                    to_io <: state;
                } else if (io_request == DUMP) {
                    to_io <: log_index;
                    for (int i=0; i<LOGSIZE; i++) {
                        to_io <: log[i];
                    }
                    dont_stop = 0;
                }
                break;
            case from_imu :> delta_ticks:
                delta_t = (float) delta_ticks / TICKS_PER_SECOND;
                state.elapsed += delta_t;
                from_imu :> estimate;

                // low-pass filter the angle estimate
                angle_buffer[buffer_index] = estimate - OFFSET;
                buffer_index = (buffer_index + 1) % BUFFER_LENGTH;
                accumulator = 0.0;
                for(int i=0;  i < BUFFER_LENGTH; i++) {
                    accumulator += angle_buffer[i];
                }
                state.estimate = accumulator / BUFFER_LENGTH;
                //if((fabs(state.estimate) > MAX_ANGLE) || error_flag) {
                //    state.response = 0;
                //    error_flag = 1;
                //} else {
                    state.response = pid(pstate, state.estimate, delta_t);
                //}
                if (log_index < LOGSIZE) {
                    log[log_index++] = state;
                }
                to_leds <: (int)state.estimate;
                break;
            case from_motor :> int:
                to_motor <: state.response;
                dont_stop = !error_flag;
                break;
        }
    }
}

void io(chanend from_controller, chanend to_controller, chanend from_uart) {
    timer tmr;
    unsigned t;
    int count;
    const unsigned REPORT_INTERVAL = 0.1 * TICKS_PER_SECOND;
    robot_state state;
    int get_dump = GET_DUMP;

    console("\f     Balancing Robot\n\n");
    console("     t    angle  response\n");
    console("-------------------------\n");

    tmr :> t;
    while(1) {
        tmr when timerafter(t + REPORT_INTERVAL) :> void;
        t += REPORT_INTERVAL;
        to_controller <: UPDATE;
        from_controller :> state;
        console("%6.1f    %5.1f       %3d       \r", state.elapsed, state.estimate, state.response);
        if(get_dump && state.elapsed > (float)LOGSIZE * IMU_INTERVAL / TICKS_PER_SECOND) {
            to_controller <: DUMP;
            from_controller :> count;
            for (int i=0; i<count; i++){
                from_controller :> state;
                console("%8.3f\t%8.3f\t%3d\n", state.elapsed, state.estimate, state.response);
            }
            get_dump = 0;
        }
    }
}

int main() {
    chan from_imu;
    chan from_motor, to_motor;
    chan from_io, to_io;
    chan from_uart;
    chan to_leds;

    par {
      on tile[0] : par {
          angle_sensor(from_imu);
          controller(from_imu, from_motor, to_motor, from_io, to_io, to_leds);
          io(to_io, from_io, from_uart);
          LedManager(to_leds);
          UART(from_uart);
      }
      on tile[2] : pwm(to_motor, from_motor);
    }
    return 0;
}

